import pytest
from jwt.exceptions import (
    InvalidAudienceError,
    InvalidIssuerError,
    MissingRequiredClaimError,
)

from rivr.http import Request, Response
from rivr_jwt import JWTMiddleware


@pytest.fixture
def middleware() -> JWTMiddleware:
    return JWTMiddleware(key='secret')


@pytest.fixture
def jwt() -> str:
    return 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJuYW1lIjoiS3lsZSJ9.4tCpoxfyfjbUyLjm9_zu-r52Vxn6bFq9kp6Rt9xMs4A'


def test_process_request_sets_empty_jwt_when_none_provided(middleware):
    request = Request()
    middleware.process_request(request)

    assert request.jwt is None


def test_process_request_decodes_authorization_header(middleware, jwt):
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})
    middleware.process_request(request)

    assert request.jwt == {'name': 'Kyle'}


def test_process_request_allows_matching_audience():
    jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJhdWQiOiJleGFtcGxlLmNvbSJ9.INovSA2CyXeBwzR0Bqq-pFuxfQLVgnFpN4x1JP0Ve84'
    middleware = JWTMiddleware(key='secret', audience=['example.com'])
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})
    middleware.process_request(request)

    assert request.jwt == {'aud': 'example.com'}


def test_process_request_disallows_missing_audience(jwt):
    middleware = JWTMiddleware(key='secret', audience='prod.example.com')
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})

    with pytest.raises(MissingRequiredClaimError):
        middleware.process_request(request)

    assert request.jwt == None


def test_process_request_disallows_incorrect_audience():
    jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJhdWQiOiJleGFtcGxlLmNvbSJ9.INovSA2CyXeBwzR0Bqq-pFuxfQLVgnFpN4x1JP0Ve84'
    middleware = JWTMiddleware(key='secret', audience='prod.example.com')
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})

    with pytest.raises(InvalidAudienceError):
        middleware.process_request(request)

    assert request.jwt == None


def test_process_request_calls_401_for_invalid_token():
    jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJhdWQiOiJleGFtcGxlLmNvbSJ9.INovSA2CyXeBwzR0Bqq-pFuxfQLVgnFpN4x1JP0Ve84'
    middleware = JWTMiddleware(key='secret', audience='prod.example.com')
    middleware.custom_401 = lambda r: Response('custom 401')

    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})
    response = middleware.process_request(request)

    assert response.content == 'custom 401'
    assert request.jwt == None


def test_process_request_allows_matching_issuer():
    jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJleGFtcGxlLmNvbSJ9.c2lmFOiVCSRyegrYJjx60BzBhacHt3BZ-avr4PtGqWk'
    middleware = JWTMiddleware(key='secret', issuer='example.com')
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})
    middleware.process_request(request)

    assert request.jwt == {'iss': 'example.com'}


def test_process_request_disallows_missing_issuer(jwt):
    middleware = JWTMiddleware(key='secret', issuer='example.com')
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})

    with pytest.raises(MissingRequiredClaimError):
        middleware.process_request(request)

    assert request.jwt == None


def test_process_request_disallows_incorrect_issuer():
    jwt = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJleGFtcGxlLmNvbSJ9.c2lmFOiVCSRyegrYJjx60BzBhacHt3BZ-avr4PtGqWk'
    middleware = JWTMiddleware(key='secret', issuer='prod.example.com')
    request = Request(headers={'Authorization': 'Bearer {}'.format(jwt)})

    with pytest.raises(InvalidIssuerError):
        middleware.process_request(request)

    assert request.jwt == None


def test_process_request_decodes_cookie(middleware, jwt):
    request = Request()
    request.cookies['jwt'] = jwt
    middleware.process_request(request)

    assert request.jwt == {'name': 'Kyle'}


def test_process_request_ignores_invalid_cookie(middleware, jwt):
    request = Request()
    request.cookies['jwt'] = 'a.b.c'
    middleware.process_request(request)

    assert request.jwt is None


def test_process_response_encodes_cookie(middleware, jwt):
    response = Response()
    response.jwt_cookie = {'name': 'Kyle'}
    response = middleware.process_response(Request(), response)

    assert response.cookies['jwt'].value == jwt


def test_process_response_unsets_cookie(middleware, jwt):
    response = Response()
    response.jwt_cookie = None
    response = middleware.process_response(Request(), response)

    morsel = response.cookies['jwt']
    assert morsel.value == ''
    assert morsel['expires'] == 'Thu, 01-Jan-1970 00:00:00 GMT'



import datetime
import json
from typing import Any, Callable, Dict, Optional, Sequence, Tuple
from urllib.request import urlopen

import jwt
from jwt import InvalidTokenError, PyJWKSet
from jwt.api_jwt import decode_complete
from jwt.exceptions import PyJWTError
from rivr import Middleware
from rivr.http import Request, Response


class JWTMiddleware(Middleware):
    algorithms = ('HS256',)
    key = None

    jwks_uri = None

    custom_401: Optional[Callable[[Request], Response]] = None

    audience: Optional[Sequence[str]] = None
    issuer: Optional[str] = None

    cookie_name = 'jwt'
    cookie_secure = False

    header_name = 'AUTHORIZATION'
    header_grant_type = 'Bearer'

    def __init__(self, *args, **kwargs):
        self._jwks = None
        self._jwks_fetched_date = None

        super(JWTMiddleware, self).__init__(*args, **kwargs)

    def create_jwt(self, payload: Dict[str, Any]):
        """
        Creates a JWT from the given payload.
        """
        if not self.key:
            raise Exception('Configuration error: Missing key')

        return jwt.encode(payload, self.key, algorithm=self.algorithms[0])

    @property
    def is_jwks_expired(self) -> bool:
        assert self._jwks_fetched_date
        return (self._jwks_fetched_date + datetime.timedelta(hours=1)) < datetime.datetime.now()

    def get_jwks(self) -> PyJWKSet:
        if not self.jwks_uri:
            raise Exception('Configuration error: Missing key or jwks_uri')

        if not self._jwks or self.is_jwks_expired:
            with urlopen(self.jwks_uri) as fp:
                # FIXME perform retries
                self._jwks = PyJWKSet.from_dict(json.load(fp))
                self._jwks_fetched_date = datetime.datetime.now()

        return self._jwks

    def get_jwk(self, token: str) -> Tuple[Sequence[str], Any]:
        if self.key:
            return (self.algorithms, self.key)

        unverified = decode_complete(token, options={"verify_signature": False})
        kid = unverified['header'].get('kid')
        alg = unverified['header'].get('alg')
        if not kid:
            raise InvalidTokenError('No key ID')
        if not alg:
            raise InvalidTokenError('No alg')

        jwks = self.get_jwks()
        for key in jwks.keys:
            if key.key_id != kid:
                continue

            if key.public_key_use != 'sig':
                continue

            if alg != key._jwk_data.get('alg', None):
                continue

            return (key._jwk_data.get('alg'), key.key)

        raise InvalidTokenError('Unknown key')

    def verify_jwt(self, encoded_jwt: str) -> Dict[Any, Any]:
        """
        Verifies the given JWT token, returning the decoded object, or None.
        """

        (algorithms, key) = self.get_jwk(encoded_jwt)

        return jwt.decode(
            encoded_jwt,
            key,
            audience=self.audience,
            issuer=self.issuer,
            algorithms=algorithms,
        )

    def process_request(self, request: Request) -> Optional[Response]:
        setattr(request, 'browserid_middleware', self)
        setattr(request, 'jwt', None)

        header = request.headers[self.header_name]
        if header:
            bearer, token = header.split(' ', 1)

            try:
                setattr(request, 'jwt', self.verify_jwt(token))
            except PyJWTError:
                if self.custom_401:
                    return self.custom_401(request)
                raise
        elif self.cookie_name in request.cookies:
            cookie = request.cookies[self.cookie_name].value

            try:
                setattr(request, 'jwt', self.verify_jwt(cookie))
            except PyJWTError:
                # ignore invalid cookies
                pass

        else:
            setattr(request, 'jwt', None)

        return None

    def process_response(self, request: Request, response: Response) -> Response:
        if hasattr(response, 'jwt_cookie'):
            jwt_cookie = getattr(response, 'jwt_cookie')
            if jwt_cookie:
                encoded_jwt = self.create_jwt(jwt_cookie)
                response.set_cookie(
                    self.cookie_name,
                    encoded_jwt,
                    secure=self.cookie_secure,
                )
            else:
                response.delete_cookie(self.cookie_name)

        return response



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='rivr-jwt',
    version='0.3.0',
    description='Making authentication with JWT (JSON Web Token) and rivr simple.',
    url='https://github.com/rivrproject/rivr-jwt',
    packages=find_packages(),
    package_data={
        'rivr_jwt': ['py.typed'],
    },
    install_requires=[
        'rivr>=0.9.0',
        'PyJWT>=2.3.0',
    ],
    author='Kyle Fuller',
    author_email='kyle@fuller.li',
    license='BSD',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: BSD License',
    ],
)
